/*
 * LAMIP-0.0.3 - input plugin to read mp3 and id3tag
 * need :
 * - >=libid3-3.8.3 (older version have bug with gcc3)
 * - >=libmad-0.15b
*/

#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <stdio.h>
#include <mad.h>
#include <id3.h>
#include <id3/globals.h>

#include <plug_in.h>

//#define ACCURATE_LENGTH_EVEN_NO_XING
#define PARSE_APE2_TAG_FOR_REPLAY_GAIN

#ifdef ACCURATE_LENGTH_EVEN_NO_XING
#include <fcntl.h>
#include <unistd.h> // for close() !! TODO : remove this ! not portable !
#include <sys/mman.h>
#endif

/* max size of a decoded frame: 2 channels, 32-bit resolution, 1152 samples */
#define MAX_DECODED_FRAME_SIZE (2 * 1152 * 4)
#define BUFFER_SIZE (64 * 1024)
#define XING_MAGIC (('X'<<24)|('i'<<16)|('n'<<8)|'g')


#ifdef DEBUG
 #define MSG(a) lamip_send_message(a)
#else
 #define MSG(a)
#endif

struct		xing {
  int		flags;
  unsigned long	frames;
  unsigned long	bytes;
  unsigned char	toc[100];
  long		scale;

};

enum {
  XING_FRAMES = 0x0001,
  XING_BYTES  = 0x0002,
  XING_TOC    = 0x0004,
  XING_SCALE  = 0x0008
};

#ifdef PARSE_APE2_TAG_FOR_REPLAY_GAIN
struct APETagFooter {
  unsigned char ID[8];
  unsigned char Version[4];
  unsigned char Length[4];
  unsigned char TagCount[4];
  unsigned char Flags[4];
  unsigned char Reserved[8];
};

static unsigned long	Read_LE_Uint32(const unsigned char*);
static void		parse_apev2(FILE*, unsigned int, lamipSonginfo*);
static void		skip_id3v1(FILE*, unsigned int*);
static void		set_rg(const char*, lamipSonginfo*);
#endif

static int	mp3_init(const lamipPluginHandle*);
static void	mp3_cleanup(void);
static void	mp3_decode(lamipURL*, int);
static void	mp3_set_songinfo(lamipURL*, lamipSonginfo*);

static size_t	ID3v2_length(const unsigned char*, size_t);
static int	parse_xing(struct xing*, struct mad_bitptr, unsigned int);
static unsigned long total_frames_from_xing(const struct mad_stream*);
static inline signed int scale(mad_fixed_t, int);
static size_t	fill_output_buffer(unsigned char*, size_t, struct mad_synth*);

static InputPlugin	mp3_functions;
static struct {
  int		resolution;	// 8,16,24 or 32 bits output
}		mp3cfg;

/****************************************************************************/

InputPlugin*	lamip_input_info(void)
{
  mp3_functions.common.name	= "mpg123 MAD decoder";
  mp3_functions.common.description = "LAMIP mpeg1-layer 1/2/3 MAD decoder";
  mp3_functions.common.init	= mp3_init;
  mp3_functions.common.cleanup	= mp3_cleanup;
  mp3_functions.decode		= mp3_decode;
  mp3_functions.set_song_info	= mp3_set_songinfo;

  return(&mp3_functions);
}

static int	mp3_init(const lamipPluginHandle* plug)
{
  char*	val;

  if(!plug)
    return(0);
  MSG("mp3_init");
  lamip_set_mime_type((lamipPluginHandle*)plug, ".mp1", "audio/mpeg");
  lamip_set_mime_type((lamipPluginHandle*)plug, ".mp2", 0);
  lamip_set_mime_type((lamipPluginHandle*)plug, ".mp3", 0);

  val = lamip_cfg_get(plug, "resolution");
  if(!val){
    val = "16";
    lamip_cfg_set((lamipPluginHandle*)plug, "resolution", val);
  }
  mp3cfg.resolution = atoi(val);
  return(1);
}

static void	mp3_cleanup(void)
{
  MSG("mp3_cleanup");
}


static void	mp3_set_songinfo(lamipURL* url, lamipSonginfo* songinfo)
{
  ID3Tag*		id3tag = NULL;
  ID3TagIterator*	iter   = NULL;
  ID3Frame*		frame  = NULL;
  char*			filename;

  filename = lamip_url_getURL(url);
  lamip_songinfo_setTotalTracks(songinfo, 1);
  lamip_songinfo_setActiveTrack(songinfo, 0);
  //lamip_songinfo_setFilename(songinfo, filename);

  id3tag = ID3Tag_New();
  if(ID3Tag_Link(id3tag, filename)<0){
    free(filename);
    return;
  }
  iter = ID3Tag_CreateIterator(id3tag);
  while((frame = ID3TagIterator_GetNext(iter)) != NULL){
    ID3Field*	field = NULL;
    char*	text;

    switch(ID3Frame_GetID(frame)){
    case ID3FID_LEADARTIST:
      {
	field = ID3Frame_GetField(frame, ID3FN_TEXT);
	text = malloc(ID3Field_Size(field)+1);
	memset(text, 0, ID3Field_Size(field)+1);
	ID3Field_GetASCII(field, text, ID3Field_Size(field));
	lamip_songinfo_setArtist(songinfo, text);
	free(text);
	break;
      }
    case ID3FID_TITLE:
      {
	field = ID3Frame_GetField(frame, ID3FN_TEXT);
	text = malloc(ID3Field_Size(field)+1);
	memset(text, 0, ID3Field_Size(field)+1);
	ID3Field_GetASCII(field, text, ID3Field_Size(field));
	lamip_songinfo_setTitle(songinfo, text);
	free(text);
	break;
      }
    case ID3FID_YEAR:
      {
	field = ID3Frame_GetField(frame, ID3FN_TEXT);
	text = malloc(ID3Field_Size(field)+1);
	memset(text, 0, ID3Field_Size(field)+1);
	ID3Field_GetASCII(field, text, ID3Field_Size(field));
	lamip_songinfo_setYear(songinfo, text);
	free(text);
	break;
      }
    case ID3FID_ALBUM:
      {
	field = ID3Frame_GetField(frame, ID3FN_TEXT);
	text = malloc(ID3Field_Size(field)+1);
	memset(text, 0, ID3Field_Size(field)+1);
	ID3Field_GetASCII(field, text, ID3Field_Size(field));
	lamip_songinfo_setAlbum(songinfo, text);
	free(text);
	break;
      }
    case ID3FID_COMMENT:
      {
	field = ID3Frame_GetField(frame, ID3FN_TEXT);
	text = malloc(ID3Field_Size(field)+1);
	memset(text, 0, ID3Field_Size(field)+1);
	ID3Field_GetASCII(field, text, ID3Field_Size(field));
	free(text);
	break;
      }
    case ID3FID_TRACKNUM:
      {
	field = ID3Frame_GetField(frame, ID3FN_TEXT);
	text = malloc(ID3Field_Size(field)+1);
	memset(text, 0, ID3Field_Size(field)+1);
	ID3Field_GetASCII(field, text, ID3Field_Size(field));
	lamip_songinfo_setTrack(songinfo, text);
	free(text);
	break;
      }
    case ID3FID_CONTENTTYPE:
      {
	field = ID3Frame_GetField(frame, ID3FN_TEXT);
	text = malloc(ID3Field_Size(field)+1);
	memset(text, 0, ID3Field_Size(field)+1);
	ID3Field_GetASCII(field, text, ID3Field_Size(field));
	lamip_songinfo_setType(songinfo, text);
	free(text);
	break;
      }
    default:
      {
	printf("MP3: frame field not handle\n");
	break;
      }
    }
  }
  ID3TagIterator_Delete(iter);
  ID3Tag_Delete(id3tag);

#ifdef PARSE_APE2_TAG_FOR_REPLAY_GAIN
  set_rg(filename, songinfo);
#endif
  free(filename);
}


static size_t	ID3v2_length(const unsigned char* buffer, size_t len)
{
  if(len < 10)
    return(0);

  len = 0;
  if(strncasecmp(buffer, "ID3", 3) == 0){
    len = (buffer[6]<<21) | (buffer[7]<<14) | (buffer[8]<<7) | buffer[9];
    len+=10;
  }
  return(len);
}

static int parse_xing(struct xing* xing, struct mad_bitptr ptr, unsigned int bitlen)
{
  if(bitlen < 64 || mad_bit_read(&ptr, 32) != XING_MAGIC)
    goto fail;
  xing->flags = mad_bit_read(&ptr, 32);
  bitlen -= 64;
  if(xing->flags & XING_FRAMES){
    if(bitlen<32)
      goto fail;
    xing->frames = mad_bit_read(&ptr, 32);
    bitlen-=32;
  }
  if(xing->flags & XING_BYTES){
    if (bitlen < 32)
      goto fail;
    xing->bytes = mad_bit_read(&ptr, 32);
    bitlen -= 32;
  }
  if(xing->flags & XING_TOC){
    int i;

    if (bitlen < 800)
      goto fail;
    for(i=0; i<100; ++i)
      xing->toc[i] = mad_bit_read(&ptr, 8);
    bitlen -= 800;
  }
  if(xing->flags & XING_SCALE){
    if (bitlen < 32)
      goto fail;
    xing->scale = mad_bit_read(&ptr, 32);
    bitlen -= 32;
  }
  return(1);
 fail:
  xing->flags = 0;
  return(0);
}

static unsigned long total_frames_from_xing(const struct mad_stream *stream)
{
  struct xing xing;
  if(parse_xing(&xing, stream->anc_ptr, stream->anc_bitlen) && (xing.flags & XING_FRAMES)){
    return(xing.frames);
  }
  return(0);
}

#ifdef ACCURATE_LENGTH_EVEN_NO_XING
/* Following two functions are based on mpg321 and mad_timer from the 
   libmad distribution */
static unsigned int scan_file(const char *filename, int filesize)
{
  /* ouch, we have to decode some frame headers */
  struct mad_stream	stream;
  struct mad_header	header;
  mad_timer_t		timer;
  unsigned long		frame_count = 0;
  unsigned long		bitrate = 0;
  unsigned char*	buffer;
  int			fd;
  int			is_vbr = 0;

  if(filesize<=0)
    return(0);

  fd = open(filename, O_RDONLY);
  if(fd == -1){
    return(0);
  }

  buffer = (unsigned char *) mmap(0, filesize, PROT_READ, MAP_SHARED, fd, 0);
  if(buffer == (void *) -1){
    printf("MP3: mmap failed\n");
    return(0);
  }
  mad_stream_init(&stream);
  mad_header_init(&header);
  mad_timer_reset(&timer);
  mad_stream_buffer(&stream, buffer, filesize);
  mad_stream_skip(&stream, ID3v2_length(buffer));

  /* There are other two ways of calculating the length of an mp3:
   * 1) Constant bitrate: One frame can provide the information
   * needed: # of frames and duration. Just see how long it
   * is and do the division.
   * 2) All: Count up the frames and duration of each frames
   * by decoding each one. We do this if we've no other
   * choice, i.e. if it's a VBR file with no Xing tag.
   */
  while(1){
    if(mad_header_decode(&header, &stream) == -1){
      if(MAD_RECOVERABLE(stream.error))
	continue;
      else
	break;
    }
    frame_count++;
    if(!is_vbr){
      // Test the first 20 frames to see if this is a VBR file
      if(!(frame_count > 20)){
	if(bitrate && header.bitrate != bitrate)
	  is_vbr = 1;
	else
	  bitrate = header.bitrate;
      }else
	break;
    }
    mad_timer_add(&timer, header.duration);
  }
  if(!is_vbr){
    double time = (filesize * 8.0) / (header.bitrate); /* time in seconds */
    double timefrac = (double)time - ((long)(time));
    mad_timer_set(&timer, (long)time, (long)(timefrac*100), 100);
  }
  mad_header_finish(&header);
  mad_stream_finish(&stream);
  munmap(buffer, filesize);
  close(fd);
  return(mad_timer_count(timer, MAD_UNITS_MILLISECONDS));
}
#endif

/*
 * fixed-point number (see mad.h) => signed integer with rounding
*/
static inline signed int scale(mad_fixed_t sample, int res)
{
  /* round */
  sample += (1L << (MAD_F_FRACBITS - res));

  /* clip */
  if (sample >= MAD_F_ONE)
    sample = MAD_F_ONE - 1;
  else if (sample < -MAD_F_ONE)
    sample = -MAD_F_ONE;

  /* quantize */
  return sample >> (MAD_F_FRACBITS + 1 - res);
}

static size_t fill_output_buffer(unsigned char* buf, size_t len, struct mad_synth* synth)
{
  const mad_fixed_t *ch1, *ch2;
  unsigned int samples, pos = 0;

  ch1 = synth->pcm.samples[0];
  ch2 = synth->pcm.samples[1];

  samples = (synth->pcm.length < len / (synth->pcm.channels * 2))
    		? synth->pcm.length
		: len / (synth->pcm.channels * 2);
  while(samples--){
    int sample;
    
    /* output in signed 16-bit */
    sample = scale(*ch1++, 16);
    buf[pos++] = (sample >> 0) & 0xff;
    buf[pos++] = (sample >> 8) & 0xff;
    if(synth->pcm.channels == 2){
      sample = scale(*ch2++, 16);
      buf[pos++] = (sample >> 0) & 0xff;
      buf[pos++] = (sample >> 8) & 0xff;
    }
  }
  return(pos);
}

/*
 * In case of certain invalid input, the last frame may not get decoded.
 *
 * According to a message from Rob Leslie to mad-dev on May 29th, 2001:
 * (http://www.mars.org/mailman/public/mad-dev/2001-May/000262.html)
 *     
 *     The solution is to append enough zero bytes to your stream buffer after
 *     the last frame to satisfy the MAD_BUFFER_GUARD.
 *
 * TODO: append MAD_BUFFER_GUARD zero bytes to the end of the buffer on last frame.
 *       (need lamip_url_eof() ?)
 *       The code to refill buffer need being modified too.
 */ 
static void	mp3_decode(lamipURL* url, int sub_trk)
{
  struct stat		statbuf;
  int			filesize;
  unsigned char		buffer[BUFFER_SIZE];
  size_t		buffer_valid;
  unsigned char		output_buffer[MAX_DECODED_FRAME_SIZE];
  struct mad_stream	stream;
  struct mad_frame	frame;
  struct mad_synth	synth;
  unsigned long		total_frames = 0;
  unsigned long		frame_count = 0;
  unsigned int		songlen;
  char*			filename;

  songlen=0;
  filename = lamip_url_getURL(url);
  if(stat(filename, &statbuf) == 0)
    filesize = statbuf.st_size;
  else
    filesize = 0;
  mad_stream_init(&stream);
  mad_frame_init(&frame);
  mad_synth_init(&synth);

  buffer_valid = lamip_url_read(url, buffer, BUFFER_SIZE);
  mad_stream_skip(&stream, ID3v2_length(buffer, buffer_valid)); // jump ID3v2.x

  while(lamip_isContinue()){
    size_t len;

    if(lamip_isSeek()){
      unsigned long ms = lamip_isSeekGetAndReset();
      unsigned int file_pos = ((double) ms / songlen) * filesize;

      lamip_url_seek(url, file_pos);
      buffer_valid = 0;
      stream.error = MAD_ERROR_BUFLEN;
      stream.sync = 0;
      stream.next_frame = NULL;
      continue;
    }
    /* refill the buffer */
    if((stream.error==MAD_ERROR_BUFLEN) || (stream.buffer==NULL)){
      size_t remaining = 0;

      if(stream.next_frame == NULL){
	; /* it's the first frame or just after seeking */
      }else{
	remaining = stream.bufend - stream.next_frame;
	if(remaining > 0 && remaining < BUFFER_SIZE){
	  memmove(buffer, stream.next_frame, remaining);
	  buffer_valid = remaining;
	}else{
	  lamip_send_message("MP3: what happened?");
	  buffer_valid = 0;
	}
      }
      buffer_valid += lamip_url_read(url, buffer + buffer_valid, BUFFER_SIZE - buffer_valid);
      //fprintf(stderr, "%5.2fk => %5.2fk (%5.2fk)\n", (double) remaining / 1000, (double) buffer_valid / 1000, (double) (buffer_valid - remaining) / 1000);
      if(buffer_valid <= remaining && lamip_url_eos(url)){ // end of stream
	break;
      }
      mad_stream_buffer(&stream, buffer, buffer_valid);
      stream.error = MAD_ERROR_NONE;
    }

    if(mad_frame_decode(&frame, &stream) == -1){
      //fprintf(stderr, "MP3: frame_cout %ld, %s\n", frame_count, mad_stream_errorstr(&stream));
      if(MAD_RECOVERABLE(stream.error) || stream.error == MAD_ERROR_BUFLEN)
	continue;
      lamip_send_message("MP3: %s\n", mad_stream_errorstr(&stream));
      break;
    }
    frame_count++;
    mad_synth_frame(&synth, &frame);
    if(frame_count == 1){
      lamipPCMInfo pcmi;

      if((total_frames = total_frames_from_xing(&stream)) > 0){
	songlen = mad_timer_count(frame.header.duration,
				  MAD_UNITS_MILLISECONDS)*total_frames;
      }else{
#ifdef ACCURATE_LENGTH_EVEN_NO_XING
	if(filesize > 0){ // it's a file
	  // this is the only way to get song length, though we don't like it
	  songlen = scan_file(filename, filesize);
	}
	if (!songlen) {
#endif
	  // assume CBR
	  songlen = 1000 * ((filesize * 8.0) / frame.header.bitrate);

#ifdef ACCURATE_LENGTH_EVEN_NO_XING
	}
#endif
      }
      pcmi.format = PCM_FORMAT_S16_LE;
      pcmi.samplerate = frame.header.samplerate;
      pcmi.channels = synth.pcm.channels;
      if(!lamip_open(&pcmi, songlen))
	break;
    }
    len = fill_output_buffer(output_buffer, MAX_DECODED_FRAME_SIZE, &synth);
    lamip_writeData(output_buffer, len);
    /*
     * If there is an ID3v1 tag, mad_frame_decode() might falsely return true.
     * So break manually if we know total_frames.
     */
    if(total_frames > 0 && frame_count >= total_frames){
      lamip_drain();
      break;
    }
  }
  if(frame_count >= 1)
    lamip_close();
  mad_synth_finish(&synth);
  mad_frame_finish(&frame);
  mad_stream_finish(&stream);
  //lamip_url_close(url);
  free(filename);
  return;
}


/* 
 * the code parsing apev2 tag is based on
 * Case's Tag, which is available at
 * http://www.saunalahti.fi/~cse/files/Tag_src.zip
 * 
 * It's released under LGPL.
 */
#ifdef PARSE_APE2_TAG_FOR_REPLAY_GAIN

static unsigned long Read_LE_Uint32(const unsigned char* p)
{   
  return ((unsigned long)p[0] <<  0) |
    ((unsigned long)p[1] <<  8) |
    ((unsigned long)p[2] << 16) |
    ((unsigned long)p[3] << 24);
}

/*
 * http://www.personal.uni-jena.de/~pfk/mpp/sv8/apetag.html
 */
static void parse_apev2(FILE* fp, unsigned int offset, lamipSonginfo* si)
{
  unsigned long		TagLen;
  unsigned long		TagCount;
  unsigned char*	buff;
  unsigned char*	p;
  unsigned char*	end;
  struct APETagFooter	footer;
  int			found_a_gain = 0;
  int			found_t_gain = 0;
  int			found_a_peak = 0;
  int			found_t_peak = 0;

  /* XXX don't use this method to fill out footer */
  if(fseek(fp, offset - sizeof(footer), SEEK_SET) != 0)
    return;
  if(fread (&footer, 1, sizeof(footer), fp) != sizeof(footer))
    return;
  if(memcmp(footer.ID, "APETAGEX", sizeof(footer.ID)) != 0)
    return;
  if(Read_LE_Uint32(footer.Version) != 2000)
    return;
  TagLen = Read_LE_Uint32(footer.Length);
  if(TagLen < sizeof(footer))
    return;
  if(fseek(fp, offset - TagLen, SEEK_SET) != 0)
    return;
  if((buff = (unsigned char*)malloc(TagLen - sizeof(footer))) == NULL)
    return;
  if(fread(buff, 1, TagLen - sizeof(footer), fp) != TagLen - sizeof(footer)) {
    free (buff);
    return;
  }
  TagCount = Read_LE_Uint32(footer.TagCount);
  end = buff + TagLen - sizeof(footer);
  for(p = buff; p < end && TagCount--;) {
    unsigned long	ksize;
    unsigned long	vsize;
    unsigned long	flags;
    char		value[32];

    vsize = Read_LE_Uint32(p);
    p += 4;
    flags = Read_LE_Uint32(p);
    p += 4;
    ksize = strlen (p);
    if (ksize > 0 && vsize > 0 && !(flags & (1 << 1))) {
      if (!found_a_gain) {
	if (strcasecmp(p, "replaygain_album_gain") == 0
	    || strcasecmp(p, "rg_audiophile") == 0) {
	  found_a_gain = 1;
	  memcpy(value, p + ksize + 1, vsize);
	  value[vsize] = '\0';
	  lamip_songinfo_setField(si, "REPLAYGAIN_ALBUM_GAIN", value);
	}
      }
      if (!found_t_gain) {
	if (strcasecmp(p, "replaygain_track_gain") == 0
	    || strcasecmp(p, "rg_ratio") == 0) {
	  found_t_gain = 1;
	  memcpy(value, p + ksize + 1, vsize);
	  value[vsize] = '\0';
	  lamip_songinfo_setField(si, "REPLAYGAIN_TRACK_GAIN", value);
	}
      }
      if (!found_a_peak) {
	if (strcasecmp(p, "replaygain_album_peak") == 0) {
	  found_a_peak = 1;
	  memcpy(value, p + ksize + 1, vsize);
	  value[vsize] = '\0';
	  lamip_songinfo_setField(si, "REPLAYGAIN_ALBUM_PEAK", value);
	}
      }
      if (!found_t_peak) {
	if (strcasecmp(p, "replaygain_track_peak") == 0
	    || strcasecmp(p, "rg_peak") == 0) {
	  found_t_peak = 1;
	  memcpy(value, p + ksize + 1, vsize);
	  value[vsize] = '\0';
	  lamip_songinfo_setField(si, "REPLAYGAIN_TRACK_PEAK", value);
	}
      }
    }
    if (found_a_gain && found_t_gain && found_a_peak && found_t_peak)
      break;
    p += ksize + 1 + vsize;
  }
  free(buff);
}

static void skip_id3v1(FILE* fp, unsigned int *offset)
{
  char buf[3];
  int id3v1_size = 128;

  if (fseek(fp, *offset - id3v1_size, SEEK_SET) != 0)
    return;
  if (fread(buf, 1, 3, fp) != 3)
    return;

  if (memcmp(buf, "TAG", 3) == 0)
    *offset -= 128;
}


static void set_rg(const char *fn, lamipSonginfo *si)
{
  FILE *fp;
  unsigned int offset, toffset;

  MSG("set replaygain");
  fp = fopen(fn, "rb");
  if(!fp){
    lamip_send_message("Bad url to retreive replaygain");
    return;
  }
  fseek(fp, 0, SEEK_END);
  offset = ftell(fp);
  /* skip other format of tags */
  do{
    toffset = offset;
    skip_id3v1(fp, &offset);
  }while (toffset != offset);
  parse_apev2(fp, offset, si);
  fclose(fp);
}
#endif
